from concurrent.futures import ThreadPoolExecutor
from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
import azure.cosmos.exceptions as exceptions
from ..database.mycosmosdb import callConnect
import concurrent.futures
from typing import List, Optional

from app.models.schemas import Time, Measurement
from .utils import read_item, get_metrics
router = APIRouter()

# test companies
companies = [
    {"title": "Station A", "slug": "station-a",
        "oid": "2afd9b42-5688-44cc-9d4a-8b50cb183650", "email": "sherry@stationa.com"},
    {"title": "Prana", "slug": "prana",
        "oid": "2e4a106c-5b38-499a-8612-dede46aa024c", "email": "alon@pranasnacks.com"},
    {"title": "Hodo", "slug": "hodo",
        "oid": "6b4d016e-8752-470d-abd2-bc0788e94887", "email": "john@hodofoods.com"},
    {"title": "Evive", "slug": "evive", "oid": "3d05034e-994d-4043-a192-956dc4ce7eb4",
        "email": "laurence.payette@evivesmoothie.com"},
    {"title": "Uncle Matt's", "slug": "unclematt's",
        "oid": "b8e6f00a-d305-43da-97ca-53dee690a01e", "email": "bob@unclematts.com"},
    {"title": "Swiftly", "slug": "swiftly",
        "oid": "86af1411-981f-4240-83ec-154486341ad3", "email": "rob.gaffney@goswift.ly"},
    {"title": "Share Mobility", "slug": "share",
        "oid": "9f47ee35-9a71-48f4-9e82-791914b206c0", "email": "bdavis@ridewithshare.com"},
    {"title": "Tru Earth", "slug": "truearth",
        "oid": "2842c53f-4936-417f-a58e-af377c12f0f4", "email": "kate@tru.earth"},
    {"title": "Caboo", "slug": "caboo", "oid": "0e33d80a-71d1-4033-8297-9d421e61f0a9",
        "email": "kevin@cabooproducts.com"},
    {"title": "Fresh Prep", "slug": "freshprep",
        "oid": "b66c9456-5cbe-455e-a9e8-407a322357b0", "email": "husein@freshprep.ca"},
    {"title": "Olea", "slug": "olea",
        "oid": "cebcb440-bd03-4ac9-abb7-6135993c7f5b", "email": "tami@oleaedge.com"},
    {"title": "Holos", "slug": "holos", "oid": "fd35f80c-0c0a-40bc-8378-6e6587c1818b",
        "email": "charles@liveholos.com"},
    {"title": "Electrum", "slug": "electrum",
        "oid": "7ececd3d-e2bf-4c4a-88d9-a872c2b37bf2", "email": "max@electrum.co"}
]

companies2 = [
    { "title": "Alpha", "slug": "alpha","oid":"d0cb41f4-91aa-4fc6-885f-fff15f1a9202","email":"infoAlpha@metric-esg.com" },
    { "title": "Beta", "slug": "beta","oid":"f0e4f495-05ef-4c1a-98dc-1b33c372cf89","email":"infoBeta@metric-esg.com" },
    { "title": "Omega", "slug": "omega","oid":"ec35202f-e632-45ac-88d0-1adb1b82d659","email":"infoOmega@metric-esg.com" }
]


@router.post("/renewal-metrics")
async def get_renewal_impact_management_metrics(time: Time):
    """
    Retrieves the total values for a set of metrics for a given time period.

    Metrics:
    - co2-mitigated
    - water-conserved
    - improved-soil-quality
    - waste-reduced

    Arguments:
    - `time`: A `Time` instance specifying the year and quarter to retrieve metrics for.

    Returns:
    - A dictionary containing the total values for each of the metrics.
    """

    metric_labels = [
        "co2-mitigated",
        "water-conserved",
        "improved-soil-quality",
        "waste-reduced",
    ]

    # Initialize a dictionary to hold the metric values
    metrics = {metric: 0.0 for metric in metric_labels}

    # Define a helper function to retrieve the metrics for a single company
    def retrieve_metric(company):
        data = get_metrics(
            company["oid"], company["email"], time.year, time.quarter)
        return data

    # Use a ThreadPoolExecutor to retrieve the metrics for all companies in parallel
    with ThreadPoolExecutor(max_workers=len(companies)) as executor:
        metrics_data = executor.map(retrieve_metric, companies)

        # Aggregate the metric values for all companies
        for data in metrics_data:
            if not data:
                continue
            for metric in metric_labels:
                metric_data = next(
                    (x for x in data if x["label"] == metric), None)
                if metric_data:
                    metrics[metric] += float(metric_data["value"]["total"])

    # Return the metric values in a JSON response
    return JSONResponse(content={
        "totalCo2Mitigated": metrics["co2-mitigated"],
        "totalWaterConserved": metrics["water-conserved"],
        "totalImprovedSoilQuality": metrics["improved-soil-quality"],
        "totalWasteReduced": metrics["waste-reduced"],
    })


@router.post("/sample-metrics")
async def get_sample_impact_management_metrics(time: Time):
    metric_labels = [
        "co2-mitigated",
        "water-conserved",
        "improved-soil-quality",
        "waste-reduced",
    ]
    metrics = {metric: 0.0 for metric in metric_labels}

    def retrieve_metric(company):
        data = get_metrics(
            company["oid"], company["email"], time.year, time.quarter)
        return data

    with ThreadPoolExecutor(max_workers=len(companies2)) as executor:
        metrics_data = executor.map(retrieve_metric, companies2)
        for data in metrics_data:
            if not data:
                continue
            for metric in metric_labels:
                metric_data = next(
                    (x for x in data if x["label"] == metric), None)
                if metric_data:
                    metrics[metric] += float(metric_data["value"]["total"])

    return JSONResponse(content={
        "totalCo2Mitigated": metrics["co2-mitigated"],
        "totalWaterConserved": metrics["water-conserved"],
        "totalImprovedSoilQuality": metrics["improved-soil-quality"],
        "totalWasteReduced": metrics["waste-reduced"],
    })



@router.post("/renewal-meaurement")
async def save_renewal_impact_management_company_data(measurement: Measurement):
    """
    Save measurement data from renewal impact management page under given categories.
    metric : RenewalMeasurement

    Arguments:
    - measurement: a Pydantic model with the following fields:
        - oid: a string representing the company's object ID
        - email: a string representing the user's email address
        - year: an string representing the year for which the measurement is recorded
        - quarter: an string representing the quarter for which the measurement is recorded
        - category: a string representing the category of the measurement (accepted categories: co2-mitigated, water-conserved, improved-soil-quality, waste-reduced)
        - total: a string representing the total measurement value
        - varA: a string representing the value of a specific variable for the given category (optional)
        - varB: a string representing the value of another specific variable for the given category (optional)
        - varC: a string representing the value of another specific variable for the given category (optional)
        ..etc

    Returns:
    - a JSONResponse with the saved measurement data

    Raises:
    - HTTPException with status_code 404 if the specified data is not found in the database
    - HTTPException with status_code 400 if the specified category is not accepted
    """

    db_result = False
    try:
        db_result = read_item(callConnect.container15,
                              measurement.oid, measurement.email)
    except exceptions.CosmosHttpResponseError as e:
        raise HTTPException(status_code=404, detail="Data not found")

    metric = "RenewalMeasurement"

    # check for category
    accepted_categories = ['co2-mitigated', 'water-conserved',
                           'improved-soil-quality', 'waste-reduced']
    if measurement.category not in accepted_categories:
        raise HTTPException(status_code=400, detail="Category not accepted")

    # extracting total and varialbes
    ignore_keys = ["oid", "email", "year", "quarter", "category"]
    data = {k: v for k, v in measurement.dict(
    ).items() if v is not None and k not in ignore_keys}

    cat_met_data = {"label": measurement.category, "value": data}
    quarter_data = {"label": measurement.quarter, "metrics": [cat_met_data]}
    year_data = {"label": measurement.year, "value": [quarter_data]}
    metric_data = {"label": metric, "metrics": [year_data]}

    # check if custom measure existssss
    if not db_result or "custommeasure" not in db_result:
        return None

    # user dict
    user = {"id": db_result["id"],
            "email": db_result["email"], "custommeasure": []}
    custom_measure = db_result["custommeasure"]

    # check if renewal-measurement data exists in custom measure
    result = next(
        (x for x in custom_measure if x["label"] == metric), None)
    if result:

        other_result = [x for x in custom_measure if x["label"] != metric]

        # find data with given year
        result_year = next(
            (x for x in result["metrics"] if x["label"] == measurement.year), None)

        if result_year:
            other_result_years = [
                x for x in result["metrics"] if x["label"] != measurement.year]

            # if data with given year exists
            quarter = next(
                (x for x in result_year["value"] if x["label"] == measurement.quarter), None)

            other_quarters = [x for x in result_year["value"]
                              if x["label"] != measurement.quarter]

            if quarter:
                # if quarter exists -> update metrics
                quarter_metrics = quarter["metrics"]
                other_cat_metrics = [
                    met for met in quarter_metrics if met["label"] != measurement.category]

                cat_metrics = next(
                    (met for met in quarter_metrics if met["label"] == measurement.category), None)

                if cat_metrics:
                    for k, v in data.items():
                        cat_metrics["value"][k] = v

                    other_cat_metrics.append(cat_metrics)
                else:
                    other_cat_metrics.append(cat_met_data)
                quarter["metrics"] = other_cat_metrics
                other_quarters.append(quarter)

            else:
                # if quarter doesn't exist -> add quarter data
                other_quarters.append(quarter_data)

            result_year["value"] = other_quarters
            other_result_years.append(result_year)
        else:
            # if data with given year doesn't exist -> append year data with given quarter and metrics
            result["metrics"].append(year_data)

        user["custommeasure"] = other_result + [result]

    else:
        # append complete metric data if metric doesnot exist
        custom_measure.append(metric_data)
        user["custommeasure"] = custom_measure

    callConnect.container15.upsert_item(body=user)

    return JSONResponse(content=cat_met_data)
